/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004-2005  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    $Id: tlist.c,v 1.12 2005/05/20 08:18:18 kasperd Exp $
*/

#include "tstore.h"

/* Format string is not fully implemented and semantics
 * may change in the future as more features are added.
 */

static inline int isOdigit(int c)
{
  switch(c) {
  case '0': case '1': case '2': case '3':
  case '4': case '5': case '6': case '7':
    return 1;
  }
  return 0;
}

size_t translate_backslash(char *dest, const char *src)
{
  size_t r=0;
  while (*src) {
    if (*src == '\\') {
      switch(src[1]) {
      case '\\': dest[r++]='\\'; src+=2; break;
      case 'a' : dest[r++]='\a'; src+=2; break;
      case 'b' : dest[r++]='\b'; src+=2; break;
      case 'f' : dest[r++]='\f'; src+=2; break;
      case 'n' : dest[r++]='\n'; src+=2; break;
      case 'r' : dest[r++]='\r'; src+=2; break;
      case 't' : dest[r++]='\t'; src+=2; break;
      case 'v' : dest[r++]='\v'; src+=2; break;
      case 'x': case 'X':
	if (isxdigit(src[2])&&isxdigit(src[3])){
	  char t[3];
	  t[0]=src[2];
	  t[1]=src[3];
	  t[2]=0;
	  dest[r++]=strtol(t,NULL,16);
	  src+=4;
	} else {
	  dest[r++]=*(src++);
	}
	break;
      case '0': case '1': case '2': case '3':
      case '4': case '5': case '6': case '7':
	{
	  int i;
	  char t[4];
	  for (i=0;(isOdigit(*++src)&&(i<3));++i) t[i]=*src;
	  t[i]=0;
	  dest[r++]=strtol(t,NULL,8);
	}
	break;
      default:
	dest[r++]=*(src++);
      }
    } else {
      dest[r++]=*(src++);
    }
  }
  dest[r]=0;
  return r;
}

struct file_info {
  char *name;
  uint64_t location,newsize;
};

void print_file(uint8_t *buf, char *format, size_t formatl,
		uint64_t idx, uint64_t newsize)
{
  size_t i=0;
  while (i<formatl)
    if (format[i]=='%'){
      i+=2;
      switch(format[i-1]) {
      case '%':
	putchar('%');
	break;
      case 'a':
      case 'c':
      case 't':
	printf("%lld",GET_TIME(buf+SECONDS_OFFSET));
	break;
      case 'i':
	printf("%lld",idx);
	break;
      case 'N':
	printf("%09lld",GET_NTIM(buf+NANOSEC_OFFSET));
	break;
      case 'p':
	printf("%s",buf+FILENAME_OFFSET);
	break;
      case 's':
	printf("%lld",GET_SIZE(buf+FILESIZE_OFFSET));
	break;
      case '-':
	printf("%lld",GET_BNUM(buf+STORESIZE_OFFSET));
	break;
      case '+':
	printf("%lld",newsize-GET_BNUM(buf+STORESIZE_OFFSET));
	break;
      case '=':
	printf("%lld",newsize);
	break;
      default:
	i-=1;
	putchar('%');
	break;
      }
    } else {
      putchar(format[i++]);
    }
  fflush(stdout);
}

int main(int argc, char ** argv)
{
  uint8_t *buf;
  struct store *s;
  uint64_t idx,nidx;
  char *format="%s\t%p\n";
  size_t formatl=strlen(format);
  char *storename=NULL;
  const char *progname=argv[0];

  struct file_info *names;
  int nnames=0;
  int i;

  names=myalloc(sizeof(struct file_info)*argc);

  while (*++argv) {
    if (strcmp(*argv,"-f")) {
      if (!storename) {
	storename=*argv;
      } else {
	names[nnames].name=*argv;
	names[nnames].location=0;
	names[nnames].newsize=0;
	++nnames;
     }
    } else {
      char *t=*++argv;
      if (t) {
	format=malloc(strlen(t)+2);
	formatl=translate_backslash(format,t);
      } else {
	storename=NULL;
	break;
      }
    }
  }

  if (!storename) {
    fprintf(stderr,"Uage: %s [-f format] <store> [filenames]\n",progname);
    return 1;
  }

  s=open_store(storename,O_RDONLY,DONT_LOCK);
  if (!s) errstop(storename);
  buf=s->header_buf;

  for (idx=GET_BNUM(buf+NEXTPTR_OFFSET); idx; idx=nidx) {
    uint64_t newsize=GET_BNUM(buf+STORESIZE_OFFSET);
    get_block(s,idx,buf);
    if (nnames) {
      for (i=0;i<nnames;++i)
	if (!strcmp(names[i].name,buf+FILENAME_OFFSET)) {
	  if (names[i].location) {
	    fprintf(stderr,"%s: duplicate %s",progname,names[i].name);
	  } else {
	    names[i].location=idx;
	    names[i].newsize=newsize;
	  }
	}
    } else {
      print_file(buf,format,formatl,idx,newsize);
    }
    nidx=GET_BNUM(buf+NEXTPTR_OFFSET);
    if ((nidx>=s->store_size-1)||(nidx==idx)) {
      fprintf(stderr,"Invalid link %lld=>%lld\n",idx,nidx);
      return 1;
    }
  }
  for (i=0;i<nnames;++i)
    if (names[i].location) {
      get_block(s,names[i].location,buf);
      print_file(buf,format,formatl,names[i].location,names[i].newsize);
    } else {
      fprintf(stderr,"%s: %s not found in %s\n",progname,names[i].name,storename);
    }
  return 0;
}
